import asyncio
import json
import ssl
from enum import IntEnum

import httpx
import websockets

from .logger import logger
from .router import Router
from .utils import find_LCU_process, parse_LCU_cmdline, build_auth_header


__all__ = [
    "MoonBlade",
    "EventCode",
]


class MoonBlade(object):
    """
    A Python 3 asynchronous library committed to communicating with LOL server through the LCU API in a simple and flexible way .
    """

    __all__ = ["start", "stop", "request"]

    _pid = 0
    _port = 0
    _token = ""
    _http_alive = False
    _ws_alive = False
    
    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.stop()
        return True

    async def start(self):
        logger.info("Starting MoonBlade.")
        LCU_process = find_LCU_process()
        while not LCU_process:
            logger.warn("No running LCUxprocess, re-searching...")
            LCU_process = find_LCU_process()
            await asyncio.sleep(0.5)
        logger.info(f"Found LCUx process.")

        LCU_args = parse_LCU_cmdline(LCU_process.cmdline())
        self._pid = LCU_process.pid
        self._port = int(LCU_args["app-port"])
        self._token = LCU_args["remoting-auth-token"]
        
        await self._start_http()

        await self._start_ws()

        #  Fake subscription message to prompt MoonBlade is fully started.
        await Router.fake(None, "Update", "/moonblade/start")

        logger.info("MoonBlade is fully started.")
        return

    async def _start_http(self):
        if self._http_alive:
            logger.warn("Http is already run.")
            return
        
        headers = {"Content-Type": "application/json", "Accept": "application/json"}
        self._http_client = httpx.AsyncClient(
            base_url=f"https://127.0.0.1:{self._port}",
            auth=httpx.BasicAuth("riot", self._token),
            proxies={"all://":None},
            headers=headers,
            verify=False
        )
        self._http_alive = True

        while True:
            try:
                res = await self._http_client.get("/riotclient/ux-state")
                if res.status_code == 200:
                    logger.info(f"Connected to LCUx http server.")
                else:
                    logger.warn(f"Connected to LCUx http server, but got an invalid response for a known uri. Status: {res.status_code}")
                break
            except httpx.ConnectError:
                logger.warn(f"Cannot connect to LCUx http server, retrying...")
                await asyncio.sleep(0.5)
                
        logger.info("Http is started.")
        return
    
    async def _start_ws(self):
        if self._ws_alive:
            logger.warn("Ws is already run.")
            return
        
        headers = [
            ("Content-Type", "application/json"),
            ("Accept", "application/json"),
            ("Accept-Encoding", "gzip, deflate"),
            ("Authorization", build_auth_header("riot", self._token))
        ]

        self._ws_client = await websockets.connect(
            uri=f"wss://127.0.0.1:{self._port}",
            ssl=ssl._create_unverified_context(),  # if "sll" is set to False, an HTTP protocol handshake that is not supported by the LCU server will be used, so set "ssl" to unverified ssl.
            extra_headers=headers,
            max_size=1024*1024*10  # set the max_size enough big to prevent Error 1009.
        )
        self._ws_alive = True

        self._ws_loop_task = asyncio.create_task(self._ws_loop())  # create ws loop.

        await self._ws_client.send(json.dumps([EventCode.SUBSCRIBE, "OnJsonApiEvent"]))  # subscribe to all json events.
        
        logger.info("Ws is started.")
        return

    async def _ws_loop(self):
        logger.debug("Ws loop is created.")
        try:
            async for msg in self._ws_client:
                if msg == "":
                    logger.info("Got websocket message containing no data.")
                else:
                    data = json.loads(msg)
                    asyncio.create_task(Router._dispatch(data[2])) # dispatch messages with Router.
        except Exception as e:
            logger.warn(f"ws loop error: {e}")
            await Router.fake(None, "Update", "/riotclient/pre-shutdown/begin")
        self._ws_alive = False
        return
    
    async def request(self, method:str, uri:str, **kwargs):
        # Need to check if the http is alive.

        if kwargs.get("data", None):
            kwargs["data"] = json.dumps(kwargs["data"])
        return await self._http_client.request(method, uri, **kwargs)

    async def stop(self):
        await self._stop_http()
        await self._stop_ws()
    
    async def _stop_http(self):
        if self._http_alive:
            await self._http_client.aclose()
            self._http_alive = False

    async def _stop_ws(self):
        if self._ws_alive:
            await self._ws_client.close()
            await self._ws_loop_task
            self._ws_alive = False


class EventCode(IntEnum):
    WELCOME = 0
    PREFIX = 1
    CALL = 2
    CALLRESULT = 3
    CALLERROR = 4
    SUBSCRIBE = 5
    UNSUBSCRIBE = 6
    PUBLISH = 7
    EVENT = 8


import asyncio
import functools
from types import MethodType
from typing import Callable, Iterable


class Node:
    """
    When registering a class method, the class in which the method is located must 
    be a subclass of the ``Node``.
    """
    def __init__(cls) -> None:
        for key in cls.__dir__():
            getattr(cls, key)


class _Handler:
    def __init__(self, f):
        self.f = f
        self.m = None
        functools.update_wrapper(self, f)

    def __get__(self, obj, objtype) -> Callable:
        if obj is None:
            self.m = None
        else:
            self.m = MethodType(self.f, obj)
        return self.m
    
    def __call__(self, *args, **kwargs):
        if self.m is None:
            return self.f(*args, **kwargs)
        else:
            return self.m(*args, **kwargs)


class Router(object):
    """
    For distributing events.

    For more detailed usage instructions, reference the library based on MoonBlade
    `https://github.com/gfk-sveyigey/Diana`
    """

    registered_uris: dict[str:list] = {}
    registered_paths: dict[str:list] = {}
    _instance = None

    def __new__(cls, *args, **kwargs):
        """
        Implement Singleton.
        """

        if cls._instance is None:
            cls._instance = object.__new__(cls, *args, **kwargs)
        return cls._instance

    @classmethod
    def register(
        cls,
        route: str = "/",
        event_type: str | Iterable[str] = "All",
        handler: Callable | None = None,
    ) -> Callable | None:
        """
        Register a asynchronous method or function as handler with the given route.

        The method can be explicitly called or called as a decorator.
        
        When decorating a method, the class which the method in should be a subclass 
        of ``Node``.

        The `route` parameter supports path or uri. Register the route as a path 
        if it is end with "/" else as a uri.

        The handler needs to be able to accept the `data` parameter.

        :param route: The route string.
        :param event_type: Case insensitive. `Create`, `Update`, `Delete`, `All`,
        or an Iterable containing several parameters like `(Create, Update)`.
        :param handler: When explicitly calling the method, handler is needed, 
        and when the function is called as a decorator, handler need to be ``None``.
        :return: None

        Example:
        .. code-block:: python
            @Router.register("/lol-chat/v1/me", "All")
            async def on_chat_event(data):
                return

            Router.register("/lol-chat/v1/me", "All", handler)

        """

        if route == "":
            raise ValueError("Route cannot be empty string.")
        elif not route.startswith('/'):
            raise ValueError("Route must start with \"/\"")
        
        if len(event_type) == 0:
            event_type = "All"
        if "All" in event_type:
            event_type = ("Create", "Update", "Delete")

        if isinstance(event_type, str):
            event_type = (event_type.capitalize(), )
        else:
            event_type = [i.capitalize() for i in event_type]

        def decorator(handler_: Callable) -> Callable:
            if not asyncio.iscoroutinefunction(handler_):
                raise TypeError("The handler must be asynchronous.")
            
            if not hasattr(handler_, "__func__"):
                handler_ = _Handler(handler_)

            registered = cls.registered_paths if route.endswith("/") else cls.registered_uris

            if route not in registered.keys():
                registered[route] =[]
            event = {"handler": handler_, "event_type": event_type}
            registered[route].append(event)
            return handler_
        
        if handler is None:
            return decorator
        else:
            return decorator(handler)

    @classmethod
    async def fake(cls, data, event_type: str, uri: str):
        """
        Fake a message published by a server.

        :param data: The data dict or None.
        :param event_type: The EventType string, possible values: ``Create``, 
        ``Update``, ``Delete``.
        :param uri: The uri string.
        :return: None

        Example:
        .. code-block:: python
            await Router.fake(
                data = None, 
                event_type = "Update", 
                uri = "/lol-chat/v1/me"
            )
        """

        data = {"data": data, "eventType": event_type, "uri": uri}
        await Router._dispatch(data)

    @classmethod
    async def _dispatch(cls, data: dict):

        if data is None:
            return
        uri: str = data["uri"]
        event_type: str = data["eventType"].capitalize()

        if uri in cls.registered_uris.keys():
            for event in cls.registered_uris[uri]:
                if event_type in event["event_type"]:
                    handler = event["handler"]
                    await handler(data)

        for path in cls.registered_paths.keys():
            if uri.startswith(path):
                for event in cls.registered_paths[path]:
                    if event_type in event["event_type"]:
                        handler = event["handler"]
                        await handler(data)
        return



from .moonblade import MoonBlade, EventCode
from .router import Router, Node

__title__ = "moonblade"
__description__ = "A Python 3 asynchronous library committed to communicating with LOL server through the LCU API in a simple and flexible way ."
__version__ = "1.0.4"
__author__ = "gfk-sveyigey"
__author_email__ = "gfk_sveyigey@163.com"
__url__ = "https://github.com/gfk-sveyigey/moonblade"


__all__ = [
    "__title__",
    "__description__",
    "__version__",
    "MoonBlade",
    "EventCode",
    "Router",
    "Node",
]


import pathlib
from setuptools import setup


path = pathlib.Path(__file__).parent

long_description = (path / "README.md").read_text(encoding="utf-8")

setup(
    name = "moonblade",
    version = "1.0.4",
    description = "A Python 3 asynchronous library committed to communicating with LOL server through the LCU API in a simple and flexible way .",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/gfk-sveyigey/moonblade",
    author = "gfk-sveyigey",
    author_email = "gfk_sveyigey@163.com",
    license = "MIT",
    classifiers = [
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Microsoft :: Windows :: Windows 10",
        "Operating System :: Microsoft :: Windows :: Windows 11",
        'Programming Language :: Python :: 3',
        "Programming Language :: Python :: 3 :: Only",
    ],
    packages = ["moonblade"],
    install_requires = [
        "httpx", "psutil", "websockets"
    ],
)